# -*- coding: utf-8 -*-
#
#  output.py
#  This file is part of FORD.
#
#  Copyright 2015 Christopher MacMackin <cmacmackin@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

import errno
import sys
import os
import shutil
import time
import traceback
from itertools import chain

import jinja2

if sys.version_info[0] > 2:
    jinja2.utils.Cycler.next = jinja2.utils.Cycler.__next__
from tqdm import tqdm

import ford.sourceform
import ford.tipue_search
import ford.utils
from ford.graphmanager import GraphManager
from ford.graphs import graphviz_installed

loc = os.path.dirname(__file__)
env = jinja2.Environment(loader=jinja2.FileSystemLoader(os.path.join(loc, "templates")))
env.globals["path"] = os.path  # this lets us call path.* in templates


class Documentation(object):
    """
    Represents and handles the creation of the documentation files from
    a project.
    """

    def __init__(self, data, proj_docs, project, pagetree):
        env.globals[
            "projectData"
        ] = data  # This lets us use meta data anywhere within the template.
        # Also, in future for other template, we may not need to
        # pass the data obj.
        self.project = project
        self.data = data
        self.pagetree = []
        self.lists = []
        self.docs = []
        self.njobs = int(self.data["parallel"])
        self.parallel = self.njobs > 0

        self.index = IndexPage(data, project, proj_docs)
        self.search = SearchPage(data, project)
        if not graphviz_installed and data["graph"].lower() == "true":
            print(
                "Warning: Will not be able to generate graphs. Graphviz not installed."
            )
        if self.data["relative"]:
            graphparent = "../"
        else:
            graphparent = ""
        print("Creating HTML documentation...")
        try:
            if data["incl_src"] == "true":
                for item in project.allfiles:
                    self.docs.append(FilePage(data, project, item))
            for item in project.types:
                self.docs.append(TypePage(data, project, item))
            for item in project.absinterfaces:
                self.docs.append(AbsIntPage(data, project, item))
            for item in project.procedures:
                self.docs.append(ProcPage(data, project, item))
            for item in project.submodprocedures:
                self.docs.append(ProcPage(data, project, item))
            for item in project.modules:
                self.docs.append(ModulePage(data, project, item))
            for item in project.submodules:
                self.docs.append(ModulePage(data, project, item))
            for item in project.programs:
                self.docs.append(ProgPage(data, project, item))
            for item in project.blockdata:
                self.docs.append(BlockPage(data, project, item))
            if len(project.procedures) > 0:
                self.lists.append(ProcList(data, project))
            if data["incl_src"] == "true":
                if len(project.files) + len(project.extra_files) > 1:
                    self.lists.append(FileList(data, project))
            if len(project.modules) + len(project.submodules) > 0:
                self.lists.append(ModList(data, project))
            if len(project.programs) > 1:
                self.lists.append(ProgList(data, project))
            if len(project.types) > 0:
                self.lists.append(TypeList(data, project))
            if len(project.absinterfaces) > 0:
                self.lists.append(AbsIntList(data, project))
            if len(project.blockdata) > 1:
                self.lists.append(BlockList(data, project))
            if pagetree:
                for item in pagetree:
                    self.pagetree.append(PagetreePage(data, project, item))
        except Exception:
            if data["dbg"]:
                traceback.print_exc()
                sys.exit("Error encountered.")
            else:
                sys.exit('Error encountered. Run with "--debug" flag for traceback.')
        if graphviz_installed and data["graph"].lower() == "true":
            print("Generating graphs...")
            self.graphs = GraphManager(
                self.data["project_url"],
                self.data["output_dir"],
                self.data.get("graph_dir", ""),
                graphparent,
                self.data["coloured_edges"].lower() == "true",
            )
            for item in project.types:
                self.graphs.register(item)
            for item in project.procedures:
                self.graphs.register(item)
            for item in project.submodprocedures:
                self.graphs.register(item)
            for item in project.modules:
                self.graphs.register(item)
            for item in project.submodules:
                self.graphs.register(item)
            for item in project.programs:
                self.graphs.register(item)
            for item in project.files:
                self.graphs.register(item)
            for item in project.blockdata:
                self.graphs.register(item)
            self.graphs.graph_all()
            project.callgraph = self.graphs.callgraph
            project.typegraph = self.graphs.typegraph
            project.usegraph = self.graphs.usegraph
            project.filegraph = self.graphs.filegraph
        else:
            self.graphs = GraphManager(
                self.data["project_url"],
                self.data["output_dir"],
                self.data.get("graph_dir", ""),
                graphparent,
                self.data["coloured_edges"].lower() == "true",
            )
            project.callgraph = ""
            project.typegraph = ""
            project.usegraph = ""
            project.filegraph = ""
        if data["search"].lower() == "true":
            print("Creating search index...")
            if data["relative"]:
                self.tipue = ford.tipue_search.Tipue_Search_JSON_Generator(
                    data["output_dir"], ""
                )
            else:
                self.tipue = ford.tipue_search.Tipue_Search_JSON_Generator(
                    data["output_dir"], data["project_url"]
                )
            self.tipue.create_node(self.index.html, "index.html", {"category": "home"})
            jobs = len(self.docs) + len(self.pagetree)
            progbar = tqdm(
                chain(iter(self.docs), iter(self.pagetree)),
                total=jobs,
                unit="",
                file=sys.stdout,
            )
            for i, p in enumerate(progbar):
                self.tipue.create_node(p.html, p.loc, p.meta)
            print("")

    def writeout(self):
        print("Writing resulting documentation.")
        out_dir = self.data["output_dir"]
        try:
            if os.path.isfile(out_dir):
                os.remove(out_dir)
            elif os.path.isdir(out_dir):
                shutil.rmtree(out_dir)
            os.makedirs(out_dir, 0o755)
        except Exception as e:
            print("Error: Could not create output directory. {}".format(e.args[0]))
        os.mkdir(os.path.join(out_dir, "lists"), 0o755)
        os.mkdir(os.path.join(out_dir, "sourcefile"), 0o755)
        os.mkdir(os.path.join(out_dir, "type"), 0o755)
        os.mkdir(os.path.join(out_dir, "proc"), 0o755)
        os.mkdir(os.path.join(out_dir, "interface"), 0o755)
        os.mkdir(os.path.join(out_dir, "module"), 0o755)
        os.mkdir(os.path.join(out_dir, "program"), 0o755)
        os.mkdir(os.path.join(out_dir, "src"), 0o755)
        os.mkdir(os.path.join(out_dir, "blockdata"), 0o755)
        copytree(os.path.join(loc, "css"), os.path.join(out_dir, "css"))
        copytree(os.path.join(loc, "fonts"), os.path.join(out_dir, "fonts"))
        copytree(os.path.join(loc, "js"), os.path.join(out_dir, "js"))
        if self.data["graph"].lower() == "true":
            self.graphs.output_graphs(self.njobs)
        if self.data["search"].lower() == "true":
            copytree(
                os.path.join(loc, "tipuesearch"), os.path.join(out_dir, "tipuesearch")
            )
            self.tipue.print_output()
        if "media_dir" in self.data:
            try:
                copytree(self.data["media_dir"], os.path.join(out_dir, "media"))
            except OSError:
                print(
                    "Warning: error copying media directory {}".format(
                        self.data["media_dir"]
                    )
                )
        if "css" in self.data:
            shutil.copy(self.data["css"], os.path.join(out_dir, "css", "user.css"))
        if self.data["favicon"] == "default-icon":
            shutil.copy(
                os.path.join(loc, "favicon.png"), os.path.join(out_dir, "favicon.png")
            )
        else:
            shutil.copy(self.data["favicon"], os.path.join(out_dir, "favicon.png"))
        if self.data["incl_src"] == "true":
            for src in self.project.allfiles:
                shutil.copy(src.path, os.path.join(out_dir, "src", src.name))
        if "mathjax_config" in self.data:
            os.mkdir(os.path.join(out_dir, "js", "MathJax-config"))
            shutil.copy(
                self.data["mathjax_config"],
                os.path.join(
                    out_dir,
                    "js",
                    "MathJax-config",
                    os.path.basename(self.data["mathjax_config"]),
                ),
            )
        # By doing this we omit a duplication of data.
        for p in self.docs:
            p.writeout()
        for p in self.lists:
            p.writeout()
        for p in self.pagetree:
            p.writeout()
        self.index.writeout()
        self.search.writeout()


class BasePage(object):
    """
    Abstract class for representation of pages in the documentation.

      data
        Dictionary containing project information (to be used when rendering)
      proj
        FortranProject object
      obj
        The object/item in the code which this page is documenting
    """

    def __init__(self, data, proj, obj=None):
        self.data = data
        self.proj = proj
        self.obj = obj
        self.meta = getattr(obj, "meta", {})

    @property
    def out_dir(self):
        """Returns the output directory of the project"""
        return self.data["output_dir"]

    @property
    def html(self):
        """Wrapper for only doing the rendering on request (drastically reduces memory)"""
        return self.render(self.data, self.proj, self.obj)

    def writeout(self):
        with open(self.outfile, "wb") as out:
            out.write(self.html.encode("utf8"))

    def render(self, data, proj, obj):
        """
        Get the HTML for the page. This method must be overridden. Arguments
        are proj_data, project object, and item in the code which the
        page documents.
        """
        raise NotImplementedError("Should not instantiate BasePage type")


class IndexPage(BasePage):
    @property
    def outfile(self):
        return os.path.join(self.out_dir, "index.html")

    def render(self, data, proj, obj):
        if data["relative"]:
            data["project_url"] = "."
            ford.sourceform.set_base_url(".")
            ford.pagetree.set_base_url(".")
        template = env.get_template("index.html")
        return template.render(data, project=proj, proj_docs=obj)


class SearchPage(BasePage):
    @property
    def outfile(self):
        return os.path.join(self.out_dir, "search.html")

    def render(self, data, proj, obj):
        if data["relative"]:
            data["project_url"] = "."
            ford.sourceform.set_base_url(".")
            ford.pagetree.set_base_url(".")
        template = env.get_template("search.html")
        return template.render(data, project=proj)


class ProcList(BasePage):
    @property
    def outfile(self):
        return os.path.join(self.out_dir, "lists", "procedures.html")

    def render(self, data, proj, obj):
        if data["relative"]:
            data["project_url"] = ".."
            ford.sourceform.set_base_url("..")
            ford.pagetree.set_base_url("..")
        template = env.get_template("proc_list.html")
        return template.render(data, project=proj)


class FileList(BasePage):
    @property
    def outfile(self):
        return os.path.join(self.out_dir, "lists", "files.html")

    def render(self, data, proj, obj):
        if data["relative"]:
            data["project_url"] = ".."
            ford.sourceform.set_base_url("..")
            ford.pagetree.set_base_url("..")
        template = env.get_template("file_list.html")
        return template.render(data, project=proj)


class ModList(BasePage):
    @property
    def outfile(self):
        return os.path.join(self.out_dir, "lists", "modules.html")

    def render(self, data, proj, obj):
        if data["relative"]:
            data["project_url"] = ".."
            ford.sourceform.set_base_url("..")
            ford.pagetree.set_base_url("..")
        template = env.get_template("mod_list.html")
        return template.render(data, project=proj)


class ProgList(BasePage):
    @property
    def outfile(self):
        return os.path.join(self.out_dir, "lists", "programs.html")

    def render(self, data, proj, obj):
        if data["relative"]:
            data["project_url"] = ".."
            ford.sourceform.set_base_url("..")
            ford.pagetree.set_base_url("..")
        template = env.get_template("prog_list.html")
        return template.render(data, project=proj)


class TypeList(BasePage):
    @property
    def outfile(self):
        return os.path.join(self.out_dir, "lists", "types.html")

    def render(self, data, proj, obj):
        if data["relative"]:
            data["project_url"] = ".."
            ford.sourceform.set_base_url("..")
            ford.pagetree.set_base_url("..")
        template = env.get_template("types_list.html")
        return template.render(data, project=proj)


class AbsIntList(BasePage):
    @property
    def outfile(self):
        return os.path.join(self.out_dir, "lists", "absint.html")

    def render(self, data, proj, obj):
        if data["relative"]:
            data["project_url"] = ".."
            ford.sourceform.set_base_url("..")
            ford.pagetree.set_base_url("..")
        template = env.get_template("absint_list.html")
        return template.render(data, project=proj)


class BlockList(BasePage):
    @property
    def outfile(self):
        return os.path.join(self.out_dir, "lists", "blockdata.html")

    def render(self, data, proj, obj):
        if data["relative"]:
            data["project_url"] = ".."
            ford.sourceform.set_base_url("..")
            ford.pagetree.set_base_url("..")
        template = env.get_template("block_list.html")
        return template.render(data, project=proj)


class DocPage(BasePage):
    """
    Abstract class to be inherited by all pages for items in the code.
    """

    @property
    def loc(self):
        return self.obj.get_dir() + "/" + self.obj.ident + ".html"

    @property
    def outfile(self):
        return os.path.join(self.out_dir, self.obj.get_dir(), self.obj.ident + ".html")


class FilePage(DocPage):
    def render(self, data, proj, obj):
        if data["relative"]:
            data["project_url"] = ".."
            ford.sourceform.set_base_url("..")
            ford.pagetree.set_base_url("..")
        template = env.get_template("file_page.html")
        return template.render(data, src=obj, project=proj)


class TypePage(DocPage):
    def render(self, data, proj, obj):
        if data["relative"]:
            data["project_url"] = ".."
            ford.sourceform.set_base_url("..")
            ford.pagetree.set_base_url("..")
        template = env.get_template("type_page.html")
        return template.render(data, dtype=obj, project=proj)


class AbsIntPage(DocPage):
    def render(self, data, proj, obj):
        if data["relative"]:
            data["project_url"] = ".."
            ford.sourceform.set_base_url("..")
            ford.pagetree.set_base_url("..")
        template = env.get_template("nongenint_page.html")
        return template.render(data, interface=obj, project=proj)


class ProcPage(DocPage):
    def render(self, data, proj, obj):
        if data["relative"]:
            data["project_url"] = ".."
            ford.sourceform.set_base_url("..")
            ford.pagetree.set_base_url("..")
        if obj.obj == "proc":
            template = env.get_template("proc_page.html")
            return template.render(data, procedure=obj, project=proj)
        else:
            if obj.generic:
                template = env.get_template("genint_page.html")
            else:
                template = env.get_template("nongenint_page.html")
            return template.render(data, interface=obj, project=proj)


class ModulePage(DocPage):
    def render(self, data, proj, obj):
        if data["relative"]:
            data["project_url"] = ".."
            ford.sourceform.set_base_url("..")
            ford.pagetree.set_base_url("..")
        template = env.get_template("mod_page.html")
        return template.render(data, module=obj, project=proj)


class ProgPage(DocPage):
    def render(self, data, proj, obj):
        if data["relative"]:
            data["project_url"] = ".."
            ford.sourceform.set_base_url("..")
            ford.pagetree.set_base_url("..")
        template = env.get_template("prog_page.html")
        return template.render(data, program=obj, project=proj)


class BlockPage(DocPage):
    def render(self, data, proj, obj):
        if data["relative"]:
            data["project_url"] = ".."
            ford.sourceform.set_base_url("..")
            ford.pagetree.set_base_url("..")
        template = env.get_template("block_page.html")
        return template.render(data, blockdat=obj, project=proj)


class PagetreePage(BasePage):
    @property
    def loc(self):
        return "page/" + self.obj.location + "/" + self.obj.filename + ".html"

    @property
    def outfile(self):
        return os.path.join(
            self.out_dir, "page", self.obj.location, self.obj.filename + ".html"
        )

    def render(self, data, proj, obj):
        if data["relative"]:
            base_url = ("../" * len(obj.hierarchy))[:-1]
            if obj.filename == "index":
                if len(obj.hierarchy) > 0:
                    base_url = base_url + "/.."
                else:
                    base_url = ".."
            ford.sourceform.set_base_url(base_url)
            ford.pagetree.set_base_url(base_url)
            data["project_url"] = base_url
        template = env.get_template("info_page.html")
        obj.contents = ford.utils.sub_links(
            ford.utils.sub_macros(ford.utils.sub_notes(obj.contents)), proj
        )
        return template.render(data, page=obj, project=proj, topnode=obj.topnode)

    def writeout(self):
        if self.obj.filename == "index":
            os.mkdir(os.path.join(self.out_dir, "page", self.obj.location), 0o755)
        super(PagetreePage, self).writeout()

        for item in self.obj.copy_subdir:
            try:
                copytree(
                    os.path.join(self.data["page_dir"], self.obj.location, item),
                    os.path.join(self.out_dir, "page", self.obj.location, item),
                )
            except Exception as e:
                print(
                    "Warning: could not copy directory {}. Error: {}".format(
                        os.path.join(self.data["page_dir"], self.obj.location, item),
                        e.args[0],
                    )
                )

        for item in self.obj.files:
            try:
                shutil.copy(
                    os.path.join(self.data["page_dir"], self.obj.location, item),
                    os.path.join(self.out_dir, "page", self.obj.location),
                )
            except Exception as e:
                print(
                    "Warning: could not copy file {}. Error: {}".format(
                        os.path.join(self.data["page_dir"], self.obj.location, item),
                        e.args[0],
                    )
                )


def copytree(src, dst):
    """Replaces shutil.copytree to avoid problems on certain file systems.

    shutil.copytree() and shutil.copystat() invoke os.setxattr(), which seems
    to fail when called for directories on at least one NFS file system.
    The current routine is a simple replacement, which should be good enough for
    Ford.
    """

    def touch(path):
        now = time.time()
        try:
            # assume it's there
            os.utime(path, (now, now))
        except os.error:
            # if it isn't, try creating the directory,
            # a file with that name
            os.makedirs(os.path.dirname(path))
            open(path, "w").close()
            os.utime(path, (now, now))

    for root, dirs, files in os.walk(src):
        relsrcdir = os.path.relpath(root, src)
        dstdir = os.path.join(dst, relsrcdir)
        if not os.path.exists(dstdir):
            try:
                os.makedirs(dstdir)
            except OSError as ex:
                if ex.errno != errno.EEXIST:
                    raise
        for ff in files:
            shutil.copy(os.path.join(root, ff), os.path.join(dstdir, ff))
            touch(os.path.join(dstdir, ff))


def truncate(string, width):
    """
    Truncates/pads the string to be the the specified length,
    including ellipsis dots if truncation occurs.
    """
    if len(string) > width:
        return string[: width - 3] + "..."
    else:
        return string.ljust(width)



# -*- coding: utf-8 -*-
#
#  graphmanager.py
#  This file is part of FORD.
#
#  Copyright 2015 Christopher MacMackin <cmacmackin@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

import os
import sys
from multiprocessing import Pool

from tqdm import tqdm

from ford.sourceform import (
    FortranFunction,
    FortranSubroutine,
    FortranInterface,
    FortranProgram,
    FortranType,
    FortranModule,
    FortranSubmoduleProcedure,
    FortranSourceFile,
    FortranBlockData,
)
import ford.graphs


def outputFuncWrap(args):
    """Wrapper function for output graphs -- needed to allow multiprocessing to
    pickle the function (must be at top level)"""

    for f in args[0:-1]:
        f.create_svg(args[-1])

    return None


class GraphManager(object):
    """
    Object which contains graphs of module use relations, type relations,
    call trees, etc. It manages these, ensures that everything that is
    needed is added at the correct time, and produces the plots for the
    list pages.

      base_url
        The URL at which the documentation will be stored. If using
        relative URLs then should be '..'.
      outdir
        The directory in which the documentation will be produced.
      graphdir:
        The location of the graphs within the output tree.
    """

    def __init__(self, base_url, outdir, graphdir, parentdir, coloured_edges):
        self.graph_objs = []
        self.modules = set()
        self.programs = set()
        self.procedures = set()
        self.types = set()
        self.sourcefiles = set()
        self.blockdata = set()
        self.graphdir = graphdir
        self.webdir = base_url + "/" + graphdir
        self.usegraph = None
        self.typegraph = None
        self.callgraph = None
        self.filegraph = None
        ford.graphs.set_coloured_edges(coloured_edges)
        ford.graphs.set_graphs_parentdir(parentdir)

    def register(self, obj):
        if obj.meta["graph"].lower() == "true":
            ford.graphs.FortranGraph.data.register(obj, type(obj))
            self.graph_objs.append(obj)

    def graph_all(self):
        for obj in tqdm(
            iter(self.graph_objs), file=sys.stdout, total=len(self.graph_objs), unit=""
        ):
            if isinstance(obj, FortranModule):
                obj.usesgraph = ford.graphs.UsesGraph(obj, self.webdir)
                obj.usedbygraph = ford.graphs.UsedByGraph(obj, self.webdir)
                self.modules.add(obj)
            elif isinstance(obj, FortranType):
                obj.inhergraph = ford.graphs.InheritsGraph(obj, self.webdir)
                obj.inherbygraph = ford.graphs.InheritedByGraph(obj, self.webdir)
                self.types.add(obj)
            elif isinstance(
                obj,
                (
                    FortranFunction,
                    FortranSubroutine,
                    FortranInterface,
                    FortranSubmoduleProcedure,
                ),
            ):
                obj.callsgraph = ford.graphs.CallsGraph(obj, self.webdir)
                obj.calledbygraph = ford.graphs.CalledByGraph(obj, self.webdir)
                obj.usesgraph = ford.graphs.UsesGraph(obj, self.webdir)
                self.procedures.add(obj)
            elif isinstance(obj, FortranProgram):
                obj.usesgraph = ford.graphs.UsesGraph(obj, self.webdir)
                obj.callsgraph = ford.graphs.CallsGraph(obj, self.webdir)
                self.programs.add(obj)
            elif isinstance(obj, FortranSourceFile):
                obj.afferentgraph = ford.graphs.AfferentGraph(obj, self.webdir)
                obj.efferentgraph = ford.graphs.EfferentGraph(obj, self.webdir)
                self.sourcefiles.add(obj)
            elif isinstance(obj, FortranBlockData):
                obj.usesgraph = ford.graphs.UsesGraph(obj, self.webdir)
                self.blockdata.add(obj)
        usenodes = list(self.modules)
        callnodes = list(self.procedures)
        for p in self.programs:
            if len(p.usesgraph.added) > 1:
                usenodes.append(p)
            if len(p.callsgraph.added) > 1:
                callnodes.append(p)
        for p in self.procedures:
            if len(p.usesgraph.added) > 1:
                usenodes.append(p)
        for b in self.blockdata:
            if len(b.usesgraph.added) > 1:
                usenodes.append(b)
        self.usegraph = ford.graphs.ModuleGraph(usenodes, self.webdir, "module~~graph")
        self.typegraph = ford.graphs.TypeGraph(self.types, self.webdir, "type~~graph")
        self.callgraph = ford.graphs.CallGraph(callnodes, self.webdir, "call~~graph")
        self.filegraph = ford.graphs.FileGraph(
            self.sourcefiles, self.webdir, "file~~graph"
        )

    def output_graphs(self, njobs=0):
        if not self.graphdir:
            return
        try:
            os.mkdir(self.graphdir, 0o755)
        except OSError:
            pass
        if njobs == 0:
            for m in self.modules:
                m.usesgraph.create_svg(self.graphdir)
                m.usedbygraph.create_svg(self.graphdir)
            for t in self.types:
                t.inhergraph.create_svg(self.graphdir)
                t.inherbygraph.create_svg(self.graphdir)
            for p in self.procedures:
                p.callsgraph.create_svg(self.graphdir)
                p.calledbygraph.create_svg(self.graphdir)
            for p in self.programs:
                p.callsgraph.create_svg(self.graphdir)
                p.usesgraph.create_svg(self.graphdir)
            for f in self.sourcefiles:
                f.afferentgraph.create_svg(self.graphdir)
                f.efferentgraph.create_svg(self.graphdir)
            for b in self.blockdata:
                b.usesgraph.create_svg(self.graphdir)
        else:
            args = []
            # args.extend([(m.usesgraph, self.graphdir) for m in self.modules])
            # args.extend([(m.usedbygraph, self.graphdir) for m in self.modules])
            # args.extend([(m.inhergraph, self.graphdir) for m in self.types])
            # args.extend([(m.inherbygraph, self.graphdir) for m in self.types])
            # args.extend([(m.callsgraph, self.graphdir) for m in self.procedures])
            # args.extend([(m.calledbygraph, self.graphdir) for m in self.procedures])
            # args.extend([(m.callsgraph, self.graphdir) for m in self.programs])
            # args.extend([(m.usesgraph, self.graphdir) for m in self.programs])
            # args.extend([(m.afferentgraph, self.graphdir) for m in self.sourcefiles])
            # args.extend([(m.efferentgraph, self.graphdir) for m in self.sourcefiles])
            # args.extend([(m.usesgraph, self.graphdir) for m in self.blockdata])

            # Note we generate all graphs for a given object in one wrapper call
            # this is to try to ensure we don't get name collisions not present
            # in the serial version (e.g. due to calling usesgraph and usedbygraph on
            # a particular module in two different processes). May not actually be needed
            # commented block above allows testing of one graph per call approach.
            args.extend(
                [(m.usesgraph, m.usedbygraph, self.graphdir) for m in self.modules]
            )
            args.extend(
                [(m.inhergraph, m.inherbygraph, self.graphdir) for m in self.types]
            )
            args.extend(
                [
                    (m.callsgraph, m.calledbygraph, self.graphdir)
                    for m in self.procedures
                ]
            )
            args.extend(
                [(m.callsgraph, m.usesgraph, self.graphdir) for m in self.programs]
            )
            args.extend(
                [
                    (m.afferentgraph, m.efferentgraph, self.graphdir)
                    for m in self.sourcefiles
                ]
            )
            args.extend([(m.usesgraph, self.graphdir) for m in self.blockdata])

            np = min(njobs, len(args))
            pool = Pool(processes=np)
            results = pool.map(outputFuncWrap, args, len(args) / np)  # noqa F841
            pool.close()
            pool.join()

        if self.usegraph:
            self.usegraph.create_svg(self.graphdir)
        if self.typegraph:
            self.typegraph.create_svg(self.graphdir)
        if self.callgraph:
            self.callgraph.create_svg(self.graphdir)
        if self.filegraph:
            self.filegraph.create_svg(self.graphdir)



#!/usr/bin/env python
# -- coding: utf-8 --
#
#  ford.py
#
#  Copyright 2014 Christopher MacMackin <cmacmackin@gmail.com>
#  This file is part of FORD.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from contextlib import contextmanager
from io import StringIO
import sys
import argparse
import markdown
import os
import subprocess
from datetime import date, datetime

import ford.fortran_project
import ford.sourceform
import ford.output
from ford.mdx_math import MathExtension
import ford.utils
import ford.pagetree

try:
    from importlib.metadata import version, PackageNotFoundError
except ModuleNotFoundError:
    from importlib_metadata import version, PackageNotFoundError
try:
    __version__ = version(__name__)
except PackageNotFoundError:
    from setuptools_scm import get_version

    __version__ = get_version(root="..", relative_to=__file__)


__appname__ = "FORD"
__author__ = "Chris MacMackin"
__credits__ = [
    "Balint Aradi",
    "Iain Barrass",
    "Izaak Beekman",
    "Jérémie Burgalat",
    "David Dickinson",
    "Gavin Huttley",
    "Harald Klimach",
    "Nick R. Papior",
    "Marco Restelli",
    "Schildkroete23",
    "Stephen J. Turnbull",
    "Jacob Williams",
    "Stefano Zhagi",
]
__license__ = "GPLv3"
__maintainer__ = "Chris MacMackin"
__status__ = "Production"


@contextmanager
def stdout_redirector(stream):
    old_stdout = sys.stdout
    sys.stdout = stream
    try:
        yield
    finally:
        sys.stdout = old_stdout


LICENSES = {
    "by": (
        '<a rel="license" href="http://creativecommons.org/licenses/by/4.0/">'
        '<img alt="Creative Commons License" style="border-width:0" src="https://i.creativecommons.org/l/by/4.0/80x15.png" /></a>'
    ),
    "by-nd": (
        '<a rel="license" href="http://creativecommons.org/licenses/by-nd/4.0/">'
        '<img alt="Creative Commons License" style="border-width:0" src="https://i.creativecommons.org/l/by-nd/4.0/80x15.png" /></a>'
    ),
    "by-sa": (
        '<a rel="license" href="http://creativecommons.org/licenses/by-sa/4.0/">'
        '<img alt="Creative Commons License" style="border-width:0" src="https://i.creativecommons.org/l/by-sa/4.0/80x15.png" /></a>'
    ),
    "by-nc": (
        '<a rel="license" href="http://creativecommons.org/licenses/by-nc/4.0/">'
        '<img alt="Creative Commons License" style="border-width:0" src="https://i.creativecommons.org/l/by-nc/4.0/80x15.png" /></a>'
    ),
    "by-nc-nd": (
        '<a rel="license" href="http://creativecommons.org/licenses/by-nc-nd/4.0/">'
        '<img alt="Creative Commons License" style="border-width:0" src="https://i.creativecommons.org/l/by-nc-nd/4.0/80x15.png" /></a>'
    ),
    "by-nc-sa": (
        '<a rel="license" href="http://creativecommons.org/licenses/by-nc-sa/4.0/">'
        '<img alt="Creative Commons License" style="border-width:0" src="https://i.creativecommons.org/l/by-nc-sa/4.0/80x15.png" /></a>'
    ),
    "gfdl": '<a rel="license" href="http://www.gnu.org/licenses/old-licenses/fdl-1.2.en.html">GNU Free Documentation License</a>',
    "opl": '<a rel="license" href="http://opencontent.org/openpub/">Open Publication License</a>',
    "pdl": '<a rel="license" href="http://www.openoffice.org/licenses/PDL.html">Public Documentation License</a>',
    "bsd": '<a rel="license" href="http://www.freebsd.org/copyright/freebsd-doc-license.html">FreeBSD Documentation License</a>',
    "isc": '<a rel="license" href="https://opensource.org/licenses/ISC">ISC (Internet Systems Consortium) License</a>',
    "mit": '<a rel="license" href="https://opensource.org/licenses/MIT">MIT</a>',
    "": "",
}


def initialize():
    """
    Method to parse and check configurations of FORD, get the project's
    global documentation, and create the Markdown reader.
    """
    try:
        import multiprocessing

        ncpus = "{0}".format(multiprocessing.cpu_count())
    except (ImportError, NotImplementedError):
        ncpus = "0"

    # Setup the command-line options and parse them.
    parser = argparse.ArgumentParser(
        "ford",
        description="Document a program or library written in modern Fortran. Any command-line options over-ride those specified in the project file.",
    )
    parser.add_argument(
        "project_file",
        help="file containing the description and settings for the project",
        type=argparse.FileType("r"),
    )
    parser.add_argument(
        "-d",
        "--src_dir",
        action="append",
        help="directories containing all source files for the project",
    )
    parser.add_argument(
        "-p",
        "--page_dir",
        help="directory containing the optional page tree describing the project",
    )
    parser.add_argument(
        "-o", "--output_dir", help="directory in which to place output files"
    )
    parser.add_argument("-s", "--css", help="custom style-sheet for the output")
    parser.add_argument(
        "-r",
        "--revision",
        dest="revision",
        help="Source code revision the project to document",
    )
    parser.add_argument(
        "--exclude",
        action="append",
        help="any files which should not be included in the documentation",
    )
    parser.add_argument(
        "--exclude_dir",
        action="append",
        help="any directories whose contents should not be included in the documentation",
    )
    parser.add_argument(
        "-e",
        "--extensions",
        action="append",
        help="extensions which should be scanned for documentation (default: f90, f95, f03, f08)",
    )
    parser.add_argument(
        "-m",
        "--macro",
        action="append",
        help="preprocessor macro (and, optionally, its value) to be applied to files in need of preprocessing.",
    )
    parser.add_argument(
        "-w",
        "--warn",
        dest="warn",
        action="store_true",
        help="display warnings for undocumented items",
    )
    parser.add_argument(
        "-f",
        "--force",
        dest="force",
        action="store_true",
        help="continue to read file if fatal errors",
    )
    parser.add_argument(
        "--no-search",
        dest="search",
        action="store_false",
        help="don't process documentation to produce a search feature",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        dest="quiet",
        action="store_true",
        help="do not print any description of progress",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version="%(prog)s version {}".format(__version__),
    )
    parser.add_argument(
        "--debug",
        dest="dbg",
        action="store_true",
        help="display traceback if fatal exception occurs and print faulty line",
    )
    parser.add_argument(
        "-I",
        "--include",
        action="append",
        help="any directories which should be searched for include files",
    )
    parser.add_argument(
        "--externalize",
        action="store_const",
        const="true",
        help="Provide information about Fortran objects in a json database for other FORD projects to refer to.",
    )
    parser.add_argument(
        "-L",
        "--external_links",
        dest="external",
        action="append",
        help=""" External projects to link to.
If an entity is not found in the sources, FORD will try to look it up in
those external projects. If those have documentation generated by FORD with
the externalize option, a link will be placed into the documentation wherever
this entity is referred to.
FORD will look in the provided paths for a modules.json file.
""",
    )
    # Get options from command-line
    args = parser.parse_args()
    # Set up Markdown reader
    md_ext = [
        "markdown.extensions.meta",
        "markdown.extensions.codehilite",
        "markdown.extensions.extra",
        MathExtension(),
        "md_environ.environ",
    ]
    md = markdown.Markdown(
        extensions=md_ext, output_format="html5", extension_configs={}
    )
    # Read in the project-file. This will contain global documentation (which
    # will appear on the homepage) as well as any information about the project
    # and settings for generating the documentation.
    proj_docs = args.project_file.read()
    md.convert(proj_docs)
    # Remake the Markdown object with settings parsed from the project_file
    if "md_base_dir" in md.Meta:
        md_base = md.Meta["md_base_dir"][0]
    else:
        md_base = os.path.dirname(args.project_file.name)
    md_ext.append("markdown_include.include")
    if "md_extensions" in md.Meta:
        md_ext.extend(md.Meta["md_extensions"])
    md = markdown.Markdown(
        extensions=md_ext,
        output_format="html5",
        extension_configs={"markdown_include.include": {"base_path": md_base}},
    )

    # Re-read the project file
    proj_docs = md.reset().convert(proj_docs)
    proj_data = md.Meta

    # Get the default options, and any over-rides, straightened out
    options = [
        "src_dir",
        "extensions",
        "fpp_extensions",
        "fixed_extensions",
        "output_dir",
        "css",
        "exclude",
        "external",
        "externalize",
        "project",
        "author",
        "author_description",
        "author_pic",
        "summary",
        "github",
        "gitlab",
        "bitbucket",
        "facebook",
        "twitter",
        "google_plus",
        "linkedin",
        "email",
        "website",
        "project_github",
        "project_gitlab",
        "project_bitbucket",
        "project_website",
        "project_download",
        "project_sourceforge",
        "project_url",
        "doc_license",
        "display",
        "hide_undoc",
        "version",
        "year",
        "docmark",
        "predocmark",
        "docmark_alt",
        "predocmark_alt",
        "media_dir",
        "favicon",
        "warn",
        "extra_vartypes",
        "page_dir",
        "privacy_policy_url",
        "terms_of_service_url",
        "incl_src",
        "force",
        "copy_subdir",
        "alias",
        "source",
        "exclude_dir",
        "macro",
        "include",
        "preprocess",
        "quiet",
        "search",
        "lower",
        "sort",
        "extra_mods",
        "dbg",
        "graph",
        "graph_maxdepth",
        "graph_maxnodes",
        "license",
        "extra_filetypes",
        "preprocessor",
        "creation_date",
        "print_creation_date",
        "proc_internals",
        "coloured_edges",
        "graph_dir",
        "gitter_sidecar",
        "mathjax_config",
        "parallel",
        "revision",
        "fixed_length_limit",
        "max_frontpage_items",
        "encoding",
    ]
    defaults = {
        "src_dir": ["./src"],
        "extensions": ["f90", "f95", "f03", "f08", "f15"],
        "fpp_extensions": ["F90", "F95", "F03", "F08", "F15", "F", "FOR"],
        "fixed_extensions": ["f", "for", "F", "FOR"],
        "output_dir": "./doc",
        "project": "Fortran Program",
        "project_url": "",
        "display": ["public", "protected"],
        "hide_undoc": "false",
        "year": date.today().year,
        "exclude": [],
        "exclude_dir": [],
        "copy_subdir": [],
        "external": [],
        "externalize": "false",
        "docmark": "!",
        "docmark_alt": "*",
        "predocmark": ">",
        "predocmark_alt": "|",
        "alias": [],
        "favicon": "default-icon",
        "extra_vartypes": [],
        "incl_src": "true",
        "source": "false",
        "macro": [],
        "include": [],
        "preprocess": "true",
        "preprocessor": "",
        "proc_internals": "false",
        "warn": "false",
        "force": "false",
        "quiet": "false",
        "search": "true",
        "lower": "false",
        "sort": "src",
        "extra_mods": [],
        "dbg": True,
        "graph": "false",
        "graph_maxdepth": "10000",
        "graph_maxnodes": "1000000000",
        "license": "",
        "doc_license": "",
        "extra_filetypes": [],
        "creation_date": "%Y-%m-%dT%H:%M:%S.%f%z",
        "print_creation_date": "false",
        "coloured_edges": "false",
        "parallel": ncpus,
        "fixed_length_limit": "true",
        "max_frontpage_items": 10,
        "encoding": "utf-8",
    }
    listopts = [
        "extensions",
        "fpp_extensions",
        "fixed_extensions",
        "display",
        "extra_vartypes",
        "src_dir",
        "exclude",
        "exclude_dir",
        "external",
        "macro",
        "include",
        "extra_mods",
        "extra_filetypes",
        "copy_subdir",
        "alias",
    ]
    # Evaluate paths relative to project file location
    if args.warn:
        args.warn = "true"
    else:
        del args.warn
    if args.force:
        args.force = "true"
    else:
        del args.force
    if args.quiet:
        args.quiet = "true"
    else:
        del args.quiet
    if not args.search:
        args.search = "false"
    else:
        del args.search
    for option in options:
        if hasattr(args, option) and getattr(args, option):
            proj_data[option] = getattr(args, option)
        elif option in proj_data:
            # Think if there is a safe  way to evaluate any expressions found in this list
            # proj_data[option] = proj_data[option]
            if option not in listopts:
                proj_data[option] = "\n".join(proj_data[option])
        elif option in defaults:
            proj_data[option] = defaults[option]
    base_dir = os.path.abspath(os.path.dirname(args.project_file.name))
    proj_data["base_dir"] = base_dir
    for var in ["src_dir", "exclude_dir", "include"]:
        if var in proj_data:
            proj_data[var] = [
                os.path.normpath(
                    os.path.join(base_dir, os.path.expanduser(os.path.expandvars(p)))
                )
                for p in proj_data[var]
            ]
    for var in [
        "page_dir",
        "output_dir",
        "graph_dir",
        "media_dir",
        "css",
        "mathjax_config",
    ]:
        if var in proj_data:
            proj_data[var] = os.path.normpath(
                os.path.join(
                    base_dir, os.path.expanduser(os.path.expandvars(proj_data[var]))
                )
            )
    if proj_data["favicon"].strip() != defaults["favicon"]:
        proj_data["favicon"] = os.path.normpath(
            os.path.join(
                base_dir, os.path.expanduser(os.path.expandvars(proj_data["favicon"]))
            )
        )
    proj_data["display"] = [item.lower() for item in proj_data["display"]]
    proj_data["incl_src"] = proj_data["incl_src"].lower()
    proj_data["creation_date"] = datetime.now().strftime(proj_data["creation_date"])
    relative = proj_data["project_url"] == ""
    proj_data["relative"] = relative
    proj_data["extensions"] += [
        ext for ext in proj_data["fpp_extensions"] if ext not in proj_data["extensions"]
    ]
    # Parse file extensions and comment characters for extra filetypes
    extdict = {}
    for ext in proj_data["extra_filetypes"]:
        sp = ext.split()
        if len(sp) < 2:
            continue
        if len(sp) == 2:
            extdict[sp[0]] = sp[1]  # (comment_char) only
        else:
            extdict[sp[0]] = (sp[1], sp[2])  # (comment_char and lexer_str)
    proj_data["extra_filetypes"] = extdict
    # Make sure no src_dir is contained within output_dir
    for projdir in proj_data["src_dir"]:
        proj_path = ford.utils.split_path(projdir)
        out_path = ford.utils.split_path(proj_data["output_dir"])
        for directory in out_path:
            if len(proj_path) == 0:
                break
            if directory == proj_path[0]:
                proj_path.remove(directory)
            else:
                break
        else:
            print(
                "Error: directory containing source-code {} a subdirectory of output directory {}.".format(
                    projdir, proj_data["output_dir"]
                )
            )
            sys.exit(1)
    # Check that none of the docmarks are the same
    if proj_data["docmark"] == proj_data["predocmark"] != "":
        print("Error: docmark and predocmark are the same.")
        sys.exit(1)
    if proj_data["docmark"] == proj_data["docmark_alt"] != "":
        print("Error: docmark and docmark_alt are the same.")
        sys.exit(1)
    if proj_data["docmark"] == proj_data["predocmark_alt"] != "":
        print("Error: docmark and predocmark_alt are the same.")
        sys.exit(1)
    if proj_data["docmark_alt"] == proj_data["predocmark"] != "":
        print("Error: docmark_alt and predocmark are the same.")
        sys.exit(1)
    if proj_data["docmark_alt"] == proj_data["predocmark_alt"] != "":
        print("Error: docmark_alt and predocmark_alt are the same.")
        sys.exit(1)
    if proj_data["predocmark"] == proj_data["predocmark_alt"] != "":
        print("Error: predocmark and predocmark_alt are the same.")
        sys.exit(1)
    # Add gitter sidecar if specified in metadata
    if "gitter_sidecar" in proj_data:
        proj_docs += """
        <script>
            ((window.gitter = {{}}).chat = {{}}).options = {{
            room: '{}'
            }};
        </script>
        <script src="https://sidecar.gitter.im/dist/sidecar.v1.js" async defer></script>
        """.format(
            proj_data["gitter_sidecar"].strip()
        )
    # Handle preprocessor:
    if proj_data["preprocess"].lower() == "true":
        if proj_data["preprocessor"]:
            preprocessor = proj_data["preprocessor"].split()
        else:
            preprocessor = ["cpp", "-traditional-cpp", "-E", "-D__GFORTRAN__"]

        # Check whether preprocessor works (reading nothing from stdin)
        try:
            with open(os.devnull) as devnull:
                subprocess.Popen(
                    preprocessor, stdin=devnull, stdout=devnull, stderr=devnull
                ).communicate()
        except OSError as ex:
            print("Warning: Testing preprocessor failed")
            print("  Preprocessor command: {}".format(preprocessor))
            print("  Exception: {}".format(ex))
            print("  -> Preprocessing turned off")
            proj_data["preprocess"] = "false"
        else:
            proj_data["preprocess"] = "true"
            proj_data["preprocessor"] = preprocessor

    # Get the correct license for project license or use value as a custom license value.
    try:
        proj_data["license"] = LICENSES[proj_data["license"].lower()]
    except KeyError:
        print(
            'Notice: license "{}" is not a recognized value, using the value as a custom license value.'.format(
                proj_data["license"]
            )
        )
    # Get the correct license for doc license(website or doc) or use value as a custom license value.
    try:
        proj_data["doc_license"] = LICENSES[proj_data["doc_license"].lower()]
    except KeyError:
        print(
            'Notice: doc_license "{}" is not a recognized value, using the value as a custom license value.'.format(
                proj_data["doc_license"]
            )
        )

    # Return project data, docs, and the Markdown reader
    md.reset()
    return (proj_data, proj_docs, md)


def main(proj_data, proj_docs, md):
    """
    Main driver of FORD.
    """
    if proj_data["relative"]:
        proj_data["project_url"] = "."
    # Parse the files in your project
    project = ford.fortran_project.Project(proj_data)
    if len(project.files) < 1:
        print(
            "Error: No source files with appropriate extension found in specified directory."
        )
        sys.exit(1)

    # Define core macros:
    ford.utils.register_macro("url = {0}".format(proj_data["project_url"]))
    ford.utils.register_macro(
        "media = {0}".format(os.path.join(proj_data["project_url"], "media"))
    )
    ford.utils.register_macro(
        "page = {0}".format(os.path.join(proj_data["project_url"], "page"))
    )

    # Register the user defined aliases:
    for alias in proj_data["alias"]:
        ford.utils.register_macro(alias)

    # Convert the documentation from Markdown to HTML. Make sure to properly
    # handle LateX and metadata.
    if proj_data["relative"]:
        project.markdown(md, "..")
    else:
        project.markdown(md, proj_data["project_url"])
    project.correlate()
    if proj_data["relative"]:
        project.make_links("..")
    else:
        project.make_links(proj_data["project_url"])

    # Convert summaries and descriptions to HTML
    if proj_data["relative"]:
        ford.sourceform.set_base_url(".")
    if "summary" in proj_data:
        proj_data["summary"] = md.convert(proj_data["summary"])
        proj_data["summary"] = ford.utils.sub_links(
            ford.utils.sub_macros(ford.utils.sub_notes(proj_data["summary"])), project
        )
    if "author_description" in proj_data:
        proj_data["author_description"] = md.convert(proj_data["author_description"])
        proj_data["author_description"] = ford.utils.sub_links(
            ford.utils.sub_macros(
                ford.utils.sub_notes(proj_data["author_description"])
            ),
            project,
        )
    proj_docs_ = ford.utils.sub_links(
        ford.utils.sub_macros(ford.utils.sub_notes(proj_docs)), project
    )
    # Process any pages
    if "page_dir" in proj_data:
        page_tree = ford.pagetree.get_page_tree(
            os.path.normpath(proj_data["page_dir"]), proj_data["copy_subdir"], md
        )
        print()
    else:
        page_tree = None
    proj_data["pages"] = page_tree

    # Produce the documentation using Jinja2. Output it to the desired location
    # and copy any files that are needed (CSS, JS, images, fonts, source files,
    # etc.)

    docs = ford.output.Documentation(proj_data, proj_docs_, project, page_tree)
    docs.writeout()

    if proj_data["externalize"].lower() == "true":
        # save FortranModules to a JSON file which then can be used
        # for external modules
        ford.utils.external(project, make=True, path=proj_data["output_dir"])

    print("")
    return 0


def run():
    proj_data, proj_docs, md = initialize()
    if proj_data["quiet"].lower() == "true":
        f = StringIO()
        with stdout_redirector(f):
            main(proj_data, proj_docs, md)
    else:
        main(proj_data, proj_docs, md)


if __name__ == "__main__":
    run()



from ford.sourceform import FortranSourceFile
from ford.fortran_project import Project

from collections import defaultdict

import pytest


@pytest.fixture
def copy_fortran_file(tmp_path):
    def copy_file(data):
        src_dir = tmp_path / "src"
        src_dir.mkdir()
        filename = src_dir / "test.f90"
        with open(filename, "w") as f:
            f.write(data)
        settings = defaultdict(str)
        settings["docmark"] = "!"
        settings["encoding"] = "utf-8"
        settings["src_dir"] = [str(src_dir)]
        settings["extensions"] = ["f90"]
        return settings

    return copy_file


def test_use_without_rename(copy_fortran_file):
    data = """\
    module module1
      use module2, only: routine_1
      implicit none
      private
      interface routine
          module procedure :: routine_1
      end interface routine
      public :: routine
    end module module1

    module module2
      implicit none
      private
      public :: routine_1
    contains
      subroutine routine_1(i)
        integer,intent(in) :: i
        write(*,*) i
      end subroutine routine_1
    end module module2
    """

    settings = copy_fortran_file(data)
    project = Project(settings)
    project.correlate()
    assert set(project.modules[0].all_procs.keys()) == {"routine_1", "routine"}


def test_use_and_rename(copy_fortran_file):
    data = """\
    module module1
      use module2, only: routine_1 => routine
      implicit none
      private
      interface routine
          module procedure :: routine_1
      end interface routine
      public :: routine
    end module module1

    module module2
      implicit none
      private
      public :: routine
    contains
      subroutine routine(i)
        integer,intent(in) :: i
        write(*,*) i
      end subroutine routine
    end module module2
    """

    settings = copy_fortran_file(data)
    project = Project(settings)
    project.correlate()
    assert set(project.modules[0].all_procs.keys()) == {"routine_1", "routine"}


def test_module_use_only_everything(copy_fortran_file):
    data = """\
    module default_access
      ! No access keyword
      integer :: int_public, int_private
      private :: int_private
      real :: real_public
      real, private :: real_private

      type :: type_public
        complex :: component_public
        complex, private :: component_private
      end type type_public

      type :: type_private
        character(len=1) :: string_public
        character(len=1), private :: string_private
      end type type_private

      private :: sub_private, func_private, type_private

    contains
      subroutine sub_public
      end subroutine sub_public

      subroutine sub_private
      end subroutine sub_private

      integer function func_public()
      end function func_public

      integer function func_private()
      end function func_private
    end module default_access

    module use_only_everything
      use default_access, only : int_public, real_public, type_public, sub_public, func_public
    end module use_only_everything
    """

    settings = copy_fortran_file(data)
    project = Project(settings)
    project.correlate()

    # Double-check we're looking at the right module
    assert project.modules[1].name == "use_only_everything"
    assert set(project.modules[1].all_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(project.modules[1].pub_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(project.modules[1].all_types.keys()) == {
        "type_public",
    }
    assert set(project.modules[1].pub_types.keys()) == {
        "type_public",
    }
    assert set(project.modules[1].all_vars.keys()) == {
        "int_public",
        "real_public",
    }
    assert set(project.modules[1].pub_vars.keys()) == {
        "int_public",
        "real_public",
    }


def test_module_use_only_everything_change_access(copy_fortran_file):
    data = """\
    module default_access
      ! No access keyword
      integer :: int_public, int_private
      private :: int_private
      real :: real_public
      real, private :: real_private

      type :: type_public
        complex :: component_public
        complex, private :: component_private
      end type type_public

      type :: type_private
        character(len=1) :: string_public
        character(len=1), private :: string_private
      end type type_private

      private :: sub_private, func_private, type_private

    contains
      subroutine sub_public
      end subroutine sub_public

      subroutine sub_private
      end subroutine sub_private

      integer function func_public()
      end function func_public

      integer function func_private()
      end function func_private
    end module default_access

    module use_only_change_access
      use default_access, only : int_public, real_public, type_public, sub_public, func_public
      private
      public :: int_public, sub_public
    end module use_only_change_access
    """

    settings = copy_fortran_file(data)
    project = Project(settings)
    project.correlate()

    # Double-check we're looking at the right module
    assert project.modules[1].name == "use_only_change_access"
    assert set(project.modules[1].all_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(project.modules[1].pub_procs.keys()) == {
        "sub_public",
    }
    assert set(project.modules[1].all_types.keys()) == {
        "type_public",
    }
    assert set(project.modules[1].pub_types.keys()) == set()
    assert set(project.modules[1].all_vars.keys()) == {
        "int_public",
        "real_public",
    }
    assert set(project.modules[1].pub_vars.keys()) == {
        "int_public",
    }


def test_module_use_everything(copy_fortran_file):
    data = """\
    module default_access
      ! No access keyword
      integer :: int_public, int_private
      private :: int_private
      real :: real_public
      real, private :: real_private

      type :: type_public
        complex :: component_public
        complex, private :: component_private
      end type type_public

      type :: type_private
        character(len=1) :: string_public
        character(len=1), private :: string_private
      end type type_private

      private :: sub_private, func_private, type_private

    contains
      subroutine sub_public
      end subroutine sub_public

      subroutine sub_private
      end subroutine sub_private

      integer function func_public()
      end function func_public

      integer function func_private()
      end function func_private
    end module default_access

    module use_everything
      use default_access
    end module use_everything
    """

    settings = copy_fortran_file(data)
    project = Project(settings)
    project.correlate()

    # Double-check we're looking at the right module
    assert project.modules[1].name == "use_everything"
    assert set(project.modules[1].all_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(project.modules[1].pub_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(project.modules[1].all_types.keys()) == {
        "type_public",
    }
    assert set(project.modules[1].pub_types.keys()) == {
        "type_public",
    }
    assert set(project.modules[1].all_vars.keys()) == {
        "int_public",
        "real_public",
    }
    assert set(project.modules[1].pub_vars.keys()) == {
        "int_public",
        "real_public",
    }


def test_module_use_everything_reexport(copy_fortran_file):
    data = """\
    module default_access
      ! No access keyword
      integer :: int_public, int_private
      private :: int_private
      real :: real_public
      real, private :: real_private

      type :: type_public
        complex :: component_public
        complex, private :: component_private
      end type type_public

      type :: type_private
        character(len=1) :: string_public
        character(len=1), private :: string_private
      end type type_private

      private :: sub_private, func_private, type_private

    contains
      subroutine sub_public
      end subroutine sub_public

      subroutine sub_private
      end subroutine sub_private

      integer function func_public()
      end function func_public

      integer function func_private()
      end function func_private
    end module default_access

    module use_everything
      use default_access
    end module use_everything

    module reexport
      use use_everything
    end module reexport
    """

    settings = copy_fortran_file(data)
    project = Project(settings)
    project.correlate()

    # Double-check we're looking at the right module
    assert project.modules[2].name == "reexport"
    assert set(project.modules[2].all_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(project.modules[2].pub_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(project.modules[2].all_types.keys()) == {
        "type_public",
    }
    assert set(project.modules[2].pub_types.keys()) == {
        "type_public",
    }
    assert set(project.modules[2].all_vars.keys()) == {
        "int_public",
        "real_public",
    }
    assert set(project.modules[2].pub_vars.keys()) == {
        "int_public",
        "real_public",
    }


def test_member_in_other_module(copy_fortran_file):
    data = """\
    module module1
      use module2, only: testInterface
      type, public, abstract :: external_type
      contains
        procedure(testInterface), nopass, deferred :: testProc
      end type external_type
    end module module1

    module module2
      type, public, abstract :: internal_type
      contains
        procedure(testInterface), nopass, deferred :: testProc
      end type internal_type

      abstract interface
        pure function testInterface(input) result(output)
          real, intent(in) :: input
          real :: output
        end function testInterface
      end interface
    end module module2
    """

    settings = copy_fortran_file(data)
    project = Project(settings)
    project.correlate()

    module1 = project.modules[0]
    module2 = project.modules[1]

    assert (
        module1.all_types["external_type"].boundprocs[0].proto
        is module2.all_types["internal_type"].boundprocs[0].proto
    )

    assert (
        module2.all_types["internal_type"].boundprocs[0].proto.name == "testInterface"
    )
